//
//	Copyright (C) 1998 Kostya Vasilyev, kostik@pobox.com
//
//	This program is free software; you can redistribute it and/or
//	modify it under the terms of the GNU General Public License
//	as published by the Free Software Foundation; either version 2
//	of the License, or (at your option) any later version.
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

#include "KvBonesData.h"

// ---------------------------------------------------------------------------
//
// class BoneCurve
//

BoneCurve::BoneCurve()
{
	Reset();
}

BoneCurve::~BoneCurve()
{
}

int	BoneCurve::GenerateCurve(Point2* pt, int maxcount)
{
	if (kind == kLinear) {
		AxAssert(maxcount >= 4);

		pt[0] = pts[0];
		pt[1] = pts[1];
		pt[2] = pts[2];
		pt[3] = pts[3];

		return 4;
	}
	else if (kind == kSmooth) {
		int count = maxcount;
		if (count > 50) count = 50;

#if 0

		// real bezier

		float t0 = (1-t) * (1-t) * (1-t);
		float t1 = 3 * t * (1-t) * (1-t);
		float t2 = 3 * t * t * (1-t);
		float t3 = t * t * t;

		pt[i].x = t0 * pts[0].x + t1 * pts[1].x + t2 * pts[2].x + t3 * pts[3].x;
		pt[i].y = t0 * pts[0].y + t1 * pts[1].y + t2 * pts[2].y + t3 * pts[3].y;

#endif

		PrepareForGetValue();

		for (int i = 0; i < count; ++ i) {
			float t = i / float(count - 1);

#if 0

#else
			// test

			float s  = (float) sin(t * 3.1415f);
			float yb = pts[0].y + t * (pts[3].y - pts[0].y);
			float yt = y1 + t * (y2 - y1);

			pt[i].x = pts[0].x + t * (pts[3].x - pts[0].x);
			pt[i].y = s * (yt - yb) + yb;
#endif
		}

		return count;
	}

	return 0;
}

TSTR BoneCurve::Format() const
{
	TSTR t;
	t.printf("%s, %.2g, %.2g, %.2g, %.2g",
		((kind == kLinear) ? __T("LINR") : __T("SMTH")),
		pts[0].x, pts[1].x, pts[2].x, pts[3].x);
	return t;
}

void BoneCurve::PrepareForGetValue()
{
	if (kind == kSmooth) {
		if (pts[2].x - pts[1].x < 1.0e-3) {
			y1 = y2 = (pts[1].y + pts[2].y) * 0.5f;
		} else {
			float scale = (pts[2].y - pts[1].y) / (pts[2].x - pts[1].x);
			y1 = pts[1].y + (pts[0].x - pts[1].x) * scale;
			y2 = pts[1].y + (pts[3].x - pts[1].x) * scale;
		}
	}
}

bool BoneCurve::GetValue(float x, float& y)
{
	if (kind == kLinear) {
		for (int i = 0; i < 3; ++ i) {
			if (pts[i].x < x && x < pts[i + 1].x) {
				y = pts[i].y + (x - pts[i].x) * (pts[i + 1].y - pts[i].y) / (pts[i + 1].x - pts[i].x);
				return true;
			}
		}
	} else if (kind == kSmooth) {
		if (x >= pts[0].x && x <= pts[3].x) {
			float t = (x - pts[0].x) / (pts[3].x - pts[0].x);

			float s  = (float) sin(t * 3.1415f);
			float yb = pts[0].y + t * (pts[3].y - pts[0].y);
			float yt = y1 + t * (y2 - y1);

			y = s * (yt - yb) + yb;

			return true;
		}
	}

	return false;
}

void BoneCurve::Load(ILoad* iload)
{
	ULONG nread;

	iload->Read(&kind, sizeof(kind), &nread);
	iload->Read(&pts, sizeof(pts), &nread);
}

void BoneCurve::Save(ISave* isave)
{
	ULONG nwrit;

	isave->Write(&kind, sizeof(kind), &nwrit);
	isave->Write(&pts, sizeof(pts), &nwrit);
}

void BoneCurve::Reset()
{
	kind = kLinear;

	pts[0].x = -0.1f;
	pts[0].y = 0.0f;

	pts[1].x = 0.0f;
	pts[1].y = 1.0f;

	pts[2].x = 1.0f;
	pts[2].y = 1.0f;

	pts[3].x = 1.1f;
	pts[3].y = 0.0f;
}

void BoneCurve::SaveState(BoneCurveState& s)
{
	s.kind = kind;
	memcpy(s.pts, pts, sizeof(pts));
}

void BoneCurve::RestoreState(BoneCurveState& s)
{
	s.kind = kind;
	memcpy(pts, s.pts, sizeof(pts));
}

// ---------------------------------------------------------------------------
//
// struct BoneInfo
//

void BoneInfo::Save(ISave* isave)
{
	ULONG nwrit;
	
	isave->Write(&p, sizeof(p), &nwrit);
	isave->Write(&q, sizeof(q), &nwrit);
	
	isave->Write(&qmp, sizeof(qmp), &nwrit);
	isave->Write(&qmplen, sizeof(qmplen), &nwrit);
	isave->Write(&bone, sizeof(bone), &nwrit);

	curve.Save(isave);
}

void BoneInfo::Load(ILoad* iload)
{
	ULONG nread;

	pNode = NULL;
	
	iload->Read(&p, sizeof(p), &nread);
	iload->Read(&q, sizeof(q), &nread);

	iload->Read(&qmp, sizeof(qmp), &nread);
	iload->Read(&qmplen, sizeof(qmplen), &nread);
	iload->Read(&bone, sizeof(bone), &nread);

	curve.Load(iload);
}

bool BoneInfo::GetPQ(TimeValue t, Point3& p, Point3& q)
{
	INode* pParent = pNode->GetParentNode();
	if (pParent != NULL) {
		Object* parentObj = pParent->EvalWorldState(t).obj;
		if (parentObj != NULL) {
			SClass_ID	scid	= parentObj->SuperClassID();
			Class_ID	cid		= parentObj->ClassID();

			if (scid == HELPER_CLASS_ID &&
				cid == Class_ID(BONE_CLASS_ID,0)) {
				Matrix3 matQ = pNode->GetNodeTM(t);
				Matrix3 matP = pParent->GetNodeTM(t);

				p		= matP.GetRow(3);
				q		= matQ.GetRow(3);

				return true;
			}
		}
	}

	return false;
}

#ifdef ENABLE_TWIST

bool	BoneInfo::GetBase(TimeValue t, Matrix3& matBase)
{
	INode* pParent = pNode->GetParentNode();
	if (pParent != NULL) {
		Object* parentObj = pParent->EvalWorldState(t).obj;
		if (parentObj != NULL) {
			SClass_ID	scid	= parentObj->SuperClassID();
			Class_ID	cid		= parentObj->ClassID();

			if (scid == HELPER_CLASS_ID &&
				cid == Class_ID(BONE_CLASS_ID,0)) {

				Matrix3 matP = pParent->GetNodeTM(t);
				matBase = matP;

				return true;
			}
		}
	}

	return false;
}

#endif


TCHAR*	BoneInfo::GetName()
{
	return pNode->GetName();
}

bool BoneInfo::IsParentOf(BoneInfo* bone)
{
	INode* parent = bone->pNode->GetParentNode();
	return parent == pNode;
}

void BoneInfo::SaveState(BoneState& s)
{
	s.p = p;
	s.q = q;
	s.qmp = qmp;
	s.qmplen = qmplen;
	s.bone = bone;

	curve.SaveState(s.curve);
}

void BoneInfo::RestoreState(BoneState& s)
{
	p = s.p;
	q = s.q;
	qmp = s.qmp;
	qmplen = s.qmplen;
	bone = s.bone;

	curve.RestoreState(s.curve);
}

// ---------------------------------------------------------------------------
//
// struct Vtx
//

void Vtx::SaveState(BoneInfo* pBoneList, int nBoneCount, VtxState& s)
{
	s.bone0 = bone0 != NULL ? bone0 - pBoneList : -1;
	s.bone1 = bone1 != NULL ? bone1 - pBoneList : -1;

	s.wght0 = wght0;
	s.wght1 = wght1;
}

void Vtx::RestoreState(BoneInfo* pBoneList, int nBoneCount, VtxState& s)
{
	bone0 = s.bone0 != -1 && s.bone0 < nBoneCount ? pBoneList + s.bone0 : NULL;
	bone1 = s.bone1 != -1 && s.bone1 < nBoneCount ? pBoneList + s.bone1 : NULL;

	wght0 = s.wght0;
	wght1 = s.wght1;
}

// ---------------------------------------------------------------------------
//
// class KvLocalModDataState
//

KvLocalModDataState::KvLocalModDataState()
{
	m_pList = NULL;
	m_nCount = 0;
}

KvLocalModDataState::~KvLocalModDataState()
{
	delete[] m_pList;
}

void KvLocalModDataState::operator=(const KvLocalModDataState& s)
{
	delete[] m_pList;

	m_pList = new Vtx[m_nCount = s.m_nCount];
	memcpy(m_pList, s.m_pList, sizeof(Vtx) * m_nCount);
}

// ---------------------------------------------------------------------------
//
// class KvLocalModData
//

KvLocalModData::KvLocalModData(int nCount)
{
	m_pList = new Vtx[nCount];
	m_nCount = nCount;
}

KvLocalModData::~KvLocalModData()
{
}

KvLocalModData::KvLocalModData(ILoad* iload, BoneInfo* pBoneList, int nBoneCount) : m_nSeed(0)
{
	ULONG nread;	
	iload->Read(&m_nCount, sizeof(m_nCount), &nread);

	m_pList = new Vtx[m_nCount];

	for (int i = 0; i < m_nCount; ++ i) {
		Vtx& vtx = m_pList[i];
		VtxState s;
		iload->Read(&s, sizeof(s), &nread);
		vtx.RestoreState(pBoneList, nBoneCount, s);
	}
}

void KvLocalModData::Save(ISave* isave, BoneInfo* pBoneList, int nBoneCount)
{
	ULONG nwrit;	
	isave->Write(&m_nCount, sizeof(m_nCount), &nwrit);

	for (int i = 0; i < m_nCount; ++ i) {
		Vtx& vtx = m_pList[i];
		VtxState s;
		vtx.SaveState(pBoneList, nBoneCount, s);
		isave->Write(&s, sizeof(s), &nwrit);
	}
}

void KvLocalModData::SaveState(KvLocalModDataState& s)
{
	s = *this;
}

void KvLocalModData::RestoreState(KvLocalModDataState& s)
{
	KvLocalModDataState::operator=(s);
}

// ---------------------------------------------------------------------------
//
// class BonePosBuffer
//

BonePosBuffer::BonePosBuffer() :  m_pBuf(NULL), m_nCount(0)
{
}

BonePosBuffer::~BonePosBuffer()
{
	delete[] m_pBuf;
}

class IKEnum : public IKEnumCallback
{
public:
	virtual void proc(Control *c, int index);
};

void IKEnum::proc(Control *c, int index)
{
	TRACE1("IKEnum::proc %d\n", index);
}

void	BonePosBuffer::CopyPos(TimeValue t, BoneInfo* pBoneList, int nCount)
{
	if (nCount != 0)
	{
		delete[] m_pBuf;
		m_pBuf = new Matrix3[nCount];
		m_nCount = nCount;

		Matrix3* m = m_pBuf;
		for (BoneInfo* p = pBoneList; p < pBoneList + nCount; ++ p)
		{
			*m++ = p->pNode->GetNodeTM(t);

#if 1	// debug
			Control* c = p->pNode->GetTMController();
			if (c != NULL)
			{
				IKEnum ike;
				c->EnumIKParams(ike);

				Matrix3 ptm = p->pNode->GetParentTM(t);
				DOFParams dofs;
				if (c->GetDOFParams(t, ptm, dofs, FALSE) && dofs.endEffector)
				{
					TRACE1("End effector at 0x%08X\n", c);
				}
			}
#endif
		}
	}
}

bool	BonePosBuffer::PastePos(TimeValue t, BoneInfo* pBoneList, int nCount)
{
	if (nCount == m_nCount)
	{
		Matrix3* m = m_pBuf;
		for (BoneInfo* p = pBoneList; p < pBoneList + nCount; ++ p)
		{
			p->pNode->SetNodeTM(t, *m++);
		}

		return true;
	}

	return false;
}

