#pragma once

//
//	Copyright (C) 1998 Kostya Vasilyev, kostik@pobox.com
//
//	This program is free software; you can redistribute it and/or
//	modify it under the terms of the GNU General Public License
//	as published by the Free Software Foundation; either version 2
//	of the License, or (at your option) any later version.
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

#include "KvBonesDefs.h"

// ---------------------------------------------------------------------------
//
// class BoneCurveWin
//

class BoneCurveWin
{
public:
	float	minx, maxx;
	float	miny, maxy;
};

// ---------------------------------------------------------------------------
//
// class BoneCurveState
//

class BoneCurveState
{
public:
	enum Kind
	{
		kLinear,
		kSmooth
	};

	Kind	kind;
	Point2	pts[4];
};


// ---------------------------------------------------------------------------
//
// class BoneCurve
//

class BoneCurve : public BoneCurveState
{
public:
	BoneCurve();
	~BoneCurve();

	int	GenerateCurve(Point2* pt, int maxcount);
	TSTR Format() const;

	void PrepareForGetValue();
	bool GetValue(float x, float& y);

	void Load(ILoad* iload);
	void Save(ISave* isave);

	void Reset();

	void SaveState(BoneCurveState& s);
	void RestoreState(BoneCurveState& s);

	float	y1, y2;
};

// ---------------------------------------------------------------------------
//
// struct BoneState
//

struct BoneState
{
	Point3			p, q;
	Point3			qmp;	// q - p
	float			qmplen;	// || q - p ||
	bool			bone;

	BoneCurveState	curve;
};

// ---------------------------------------------------------------------------
//
// struct BoneInfo
//

struct BoneInfo
{
	INode*			pNode;
	Matrix3			matp;

	Point3			p, q;
	Point3			qmp;	// q - p
	float			qmplen;	// || q - p ||
	bool			bone;

	BoneCurve		curve;

	void			Save(ISave* isave);
	void			Load(ILoad* iload);

	bool			GetPQ(TimeValue t, Point3& p, Point3& q);
	TCHAR*			GetName();

#ifdef ENABLE_TWIST
	Matrix3			matBase;
	bool			GetBase(TimeValue t, Matrix3& matBase);
#endif

	bool IsParentOf(BoneInfo* bone);

	void SaveState(BoneState& s);
	void RestoreState(BoneState& s);
};

// ---------------------------------------------------------------------------
//
// struct VtxState
//

struct VtxState
{
	int					bone0;
	int					bone1;

	float				wght0;
	float				wght1;
};

// ---------------------------------------------------------------------------
//
// struct Vtx
//

struct Vtx
{
	BoneInfo*			bone0;
	BoneInfo*			bone1;

	float				wght0;
	float				wght1;

	void				SaveState(BoneInfo* pBoneList, int nBoneCount, VtxState& s);
	void				RestoreState(BoneInfo* pBoneList, int nBoneCount, VtxState& s);
};

// ---------------------------------------------------------------------------
//
// class KvLocalModDataState
//

class KvLocalModDataState
{
public:
	int			m_nCount;
	Vtx*		m_pList;

	KvLocalModDataState();
	~KvLocalModDataState();

	void operator=(const KvLocalModDataState& s);

private:
	KvLocalModDataState(const KvLocalModDataState& );
};

// ---------------------------------------------------------------------------
//
// class KvLocalModData
//

class KvLocalModData : public KvLocalModDataState
{
public:
	KvLocalModData(int nCount);
	~KvLocalModData();

	KvLocalModData(ILoad* iload, BoneInfo* pBoneList, int nBoneCount);
	void		Save(ISave* isave, BoneInfo* pBoneList, int nBoneCount);

	void		SaveState(KvLocalModDataState& s);
	void		RestoreState(KvLocalModDataState& s);

	int			m_nSeed;
};

// ---------------------------------------------------------------------------
//
// class BonePosBuffer
//

class BonePosBuffer
{
public:
	BonePosBuffer();
	~BonePosBuffer();

	void		CopyPos(TimeValue t, BoneInfo* pBoneList, int nCount);
	bool		PastePos(TimeValue t, BoneInfo* pBoneList, int nCount);

private:
	Matrix3*	m_pBuf;
	int			m_nCount;
};


