//
//	Copyright (C) 1998 Kostya Vasilyev, kostik@pobox.com
//
//	This program is free software; you can redistribute it and/or
//	modify it under the terms of the GNU General Public License
//	as published by the Free Software Foundation; either version 2
//	of the License, or (at your option) any later version.
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

#include "KvBonesDialogs.h"
#include "KvBones.h"
#include "KvBonesData.h"

#include <objbase.h>
#include <richedit.h>

// ---------------------------------------------------------------------------
//
// Dialog helpers
//

int ListCtrl_GetSelected(HWND hWndList)
{
	int nCount = SendMessage(hWndList, LVM_GETSELECTEDCOUNT, 0, 0);
	if (nCount == 0)
		return -1;

	nCount = SendMessage(hWndList, LVM_GETITEMCOUNT, 0, 0);
	while (nCount -- > 0) {
		if (SendMessage(hWndList, LVM_GETITEMSTATE, nCount, LVIS_SELECTED) != 0)
			return nCount;
 	}

	return -1;
}

static bool EditCtrl_GetFloat(HWND hWnd, UINT nCtrlID, float& f)
{
	TCHAR buf[100];
	SendDlgItemMessage(hWnd, nCtrlID, WM_GETTEXT, sizeof(buf) / sizeof(TCHAR), (LPARAM) buf);

	if (buf[0] != __T('\0')) {
		TCHAR* endptr;
		f = (float) strtod(buf, &endptr);
		if (endptr == buf + strlen(buf))
			return true;
	}

	TCHAR msg[200];
	_snprintf(msg, sizeof(msg) / sizeof(TCHAR), "Bad number: %s", buf);
	MessageBox(hWnd, msg, __T("Error"), MB_OK | MB_ICONSTOP);
	return false;
}


// ---------------------------------------------------------------------------
//
// About dialog box
//

/////////////////////////////////////////////////////////////////////////////

struct StreamParams
{
	LPBYTE	pszText;
	LPBYTE	pszEnd;
};

static StreamParams	g_StreamParams;

DWORD CALLBACK StreamCallback(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG FAR *pcb)
{
	StreamParams* params = (StreamParams*) dwCookie;

	if (cb > params->pszEnd - params->pszText)
		cb = params->pszEnd - params->pszText;

	memcpy(pbBuff, params->pszText, cb);
	*pcb = cb;

	params->pszText += cb;

	return 0;
}

static void		SendHelpTextToControl(HWND hWnd)
{
	HWND hWndCtrl = GetDlgItem(hWnd, IDCTL_ABOUT_RTF);

	EDITSTREAM stream;
	stream.dwCookie		= (DWORD) &g_StreamParams;
	stream.dwError		= 0;
	stream.pfnCallback	= StreamCallback;

	SendMessage(hWndCtrl, EM_STREAMIN, SF_RTF, (LPARAM) &stream); 
}

BOOL CALLBACK AboutDialogProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg) {
		case WM_INITDIALOG:
			SendDlgItemMessage(hWnd, IDC_BUILD, WM_SETTEXT, 0, (LPARAM) ( __T("Version: ") __DATE__ ) );
			CenterWindow(hWnd,GetParent(hWnd));
			SendHelpTextToControl(hWnd);
			return 1;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDOK:				  
				case IDCANCEL:
					EndDialog(hWnd,1);
					break;
			}
			return TRUE;
	}

	return FALSE;
}

void DoAboutBox(HINSTANCE hInstance, HWND hWnd)
{
	g_StreamParams.pszText		= NULL;
	g_StreamParams.pszEnd		= NULL;

	HRSRC hResInfo = FindResource(hInstance, "AboutText", "RTF");
	if (hResInfo != NULL)
	{
		HGLOBAL hgResource = LoadResource(hInstance, hResInfo);
		if (hgResource != NULL)
		{
			void* pBuf = LockResource(hgResource);
			if (pBuf != NULL)
			{
				g_StreamParams.pszText	= (unsigned char*) pBuf;
				g_StreamParams.pszEnd	= g_StreamParams.pszText + SizeofResource(hInstance, hResInfo);

				::LoadLibrary(__T("RICHED32.DLL"));
				::DialogBox(hInstance, MAKEINTRESOURCE(IDD_ABOUT), hWnd, AboutDialogProc);
			}
		}
	}
}

// ---------------------------------------------------------------------------
//
// Edit Bones dialog box
//

static TSTR EditBones_FormatBoneInfo(BoneInfo* pBone)
{
	return pBone->curve.Format();
}

static void EditBones_OnInitDialog(HWND hWnd)
{
	CenterWindow(hWnd, GetParent(hWnd));

	// ---- Init the list control

	HWND hWndList = GetDlgItem(hWnd, IDC_BONE_LIST);

	// Insert the columns

	LV_COLUMN col;
	col.mask =		LVCF_FMT | LVCF_SUBITEM | LVCF_TEXT | LVCF_WIDTH;
	col.fmt =		LVCFMT_LEFT;

	col.pszText =	__T("Bone");
	col.iSubItem =	0;
	col.cx =		100;
	::SendMessage(hWndList, LVM_INSERTCOLUMN, 0, (LPARAM)&col);
	
	col.pszText =	__T("Curve");
	col.iSubItem =	1;
	col.cx =		120;
	::SendMessage(hWndList, LVM_INSERTCOLUMN, 1, (LPARAM)&col);

	// ----- Insert info about bones

	BoneInfo* pb = KvBones::g_pCurr->m_pBoneList;
	BoneInfo* pe = KvBones::g_pCurr->m_pBoneList + KvBones::g_pCurr->m_nBoneCount;

	int iItem = 0;

	while (pb != pe) {
		if (pb->bone) {
			TSTR t = EditBones_FormatBoneInfo(pb);

			TCHAR* ptcsName = pb->GetName();

			LV_ITEM item;

			item.mask =		LVIF_TEXT | LVIF_INDENT | LVIF_PARAM;
			item.iItem =	iItem ++;
			item.iSubItem =	0;
			item.pszText =	ptcsName;
			item.iIndent =	0;
			item.lParam =	pb - KvBones::g_pCurr->m_pBoneList;
			::SendMessage(hWndList, LVM_INSERTITEM, 0, (LPARAM) &item);

				item.mask =		LVIF_TEXT;
				item.iSubItem =	1;
				item.pszText =	t;
				::SendMessage(hWndList, LVM_SETITEM, 0, (LPARAM) &item);
		}

		++ pb;
	}
}

static void EditBones_OnEditCurve(HWND hWnd)
{
	HWND hWndList = GetDlgItem(hWnd, IDC_BONE_LIST);
	int nItem = ListCtrl_GetSelected(hWndList);
	if (nItem == -1) {
		MessageBox(hWnd, __T("Please select a bone first"), __T("Error"), MB_OK | MB_ICONSTOP);
	} else {
		LV_ITEM item;
		item.iItem =	nItem;
		item.iSubItem =	0;
		item.mask =		LVIF_PARAM;
		SendMessage(hWndList, LVM_GETITEM, 0, (LPARAM) &item);
		AxAssert(item.lParam != 0);

		BoneInfo* pb = KvBones::g_pCurr->m_pBoneList + item.lParam;

		BoneCurve curve = pb->curve;
		if (DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_BONE_CURVE), hWnd, BoneCurveDialogProc, (LPARAM) &curve) == IDOK) {
			pb->curve = curve;

			TSTR t = EditBones_FormatBoneInfo(pb);

			item.mask =		LVIF_TEXT;
			item.iSubItem =	1;
			item.pszText =	t;
			::SendMessage(hWndList, LVM_SETITEM, 0, (LPARAM) &item);
		}
	}
}

static void EditBones_OnNotify(HWND hWnd, int nCtrl, NMHDR* pHdr)
{
	switch (nCtrl) {
		case IDC_BONE_LIST:
			switch (pHdr->code) {
				case NM_DBLCLK:

#if 1

					EditBones_OnEditCurve(hWnd);
#else

					nItem = ListCtrl_GetSelected(pHdr->hwndFrom);
					if (nItem != -1) {
						LV_ITEM item;
						item.iItem =	nItem;
						item.iSubItem =	0;
						item.mask =		LVIF_PARAM;
						SendMessage(pHdr->hwndFrom, LVM_GETITEM, 0, (LPARAM) &item);
						AxAssert(item.lParam != 0);

						BoneInfo* pb = KvBones::g_pCurr->m_pBoneList + item.lParam;

						if (DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_BONE_PARAMS), hWnd, BoneParamDialogProc, (LPARAM) pb) == IDOK) {
							TSTR t = EditBones_FormatBoneInfo(pb);

							item.mask =		LVIF_TEXT;
							item.iSubItem =	1;
							item.pszText =	t;
							::SendMessage(pHdr->hwndFrom, LVM_SETITEM, 0, (LPARAM) &item);
						}
					}
#endif

					break;
			}
			break;
	}
}

BOOL CALLBACK EditBoneDialogProc	(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	HWND hWndList;
	int nItem;
	BoneInfo *pb, *pe;

	switch (msg) {
		case WM_INITDIALOG:
			EditBones_OnInitDialog(hWnd);
			return TRUE;

		case WM_NOTIFY:
			EditBones_OnNotify(hWnd, wParam, (NMHDR*) lParam);
			break;

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDC_APPLY:
					KvBones::g_pCurr->ForceUpdate();
					KvBones::ip->RedrawViews(KvBones::ip->GetTime());
					break;

#if 0

				case IDC_AUTO_TENDON:
					hWndList = GetDlgItem(hWnd, IDC_BONE_LIST);
					nItem = ListCtrl_GetSelected(hWndList);
					if (nItem == -1) {
						MessageBox(hWnd, __T("Please select a bone first"), __T("Error"), MB_OK | MB_ICONSTOP);
					} else {
						// find the bone
						LV_ITEM item;
						item.iItem =	nItem;
						item.iSubItem =	0;
						item.mask =		LVIF_PARAM;
						SendMessage(hWndList, LVM_GETITEM, 0, (LPARAM) &item);
						AxAssert(item.lParam != 0);

						int nBone = item.lParam;
						if (nBone == KvBones::g_pCurr->m_nBoneCount - 1) {
							MessageBox(hWnd, __T("Cannot do this with the last bone"), __T("Error"), MB_OK | MB_ICONSTOP);
						} else {
							if (DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_AUTO_TENDON), hWnd, AutoTendonDialogProc, nBone) == IDOK) {
								TSTR t;

								t = EditBones_FormatBoneInfo(KvBones::g_pCurr->m_pBoneList + nBone);

								item.mask =		LVIF_TEXT;
								item.iSubItem =	1;
								item.pszText =	t;
								::SendMessage(hWndList, LVM_SETITEM, 0, (LPARAM) &item);

								t = EditBones_FormatBoneInfo(KvBones::g_pCurr->m_pBoneList + nBone + 1);

								item.iItem ++;
								item.mask =		LVIF_TEXT;
								item.iSubItem =	1;
								item.pszText =	t;
								::SendMessage(hWndList, LVM_SETITEM, 0, (LPARAM) &item);

								KvBones::g_pCurr->ForceUpdate();
							}
						}
					}
					break;

#endif

				case IDC_RESET_PARAMS:
					hWndList = GetDlgItem(hWnd, IDC_BONE_LIST);

					pb = KvBones::g_pCurr->m_pBoneList;
					pe = KvBones::g_pCurr->m_pBoneList + KvBones::g_pCurr->m_nBoneCount;

					nItem = 0;

					while (pb != pe) {
						if (pb->bone) {
							pb->curve.Reset();

							TSTR t = EditBones_FormatBoneInfo(pb);

							TCHAR* ptcsName = pb->GetName();

							LV_ITEM item;

							item.iItem =	nItem ++;
							item.mask =		LVIF_TEXT;
							item.iSubItem =	1;
							item.pszText =	t;
							::SendMessage(hWndList, LVM_SETITEM, 0, (LPARAM) &item);
						}

						++ pb;
					}
					break;

				case IDC_EDIT_CURVE:
					EditBones_OnEditCurve(hWnd);
					break;

				case IDOK:				  
				case IDCANCEL:
					EndDialog(hWnd, LOWORD(wParam));
					break;
			}
			return TRUE;
	}

	return FALSE;
}

#if 0

// ---------------------------------------------------------------------------
//
// Bone Parametes dialog box
//

static void BoneParam_OnInitDialog(HWND hWnd, LPARAM lParam)
{
	CenterWindow(hWnd, GetParent(hWnd));

	BoneInfo* pBone = (BoneInfo*) lParam;
	AxAssert(pBone != NULL);

	SetWindowLong(hWnd,  GWL_USERDATA, lParam);

	SendDlgItemMessage(hWnd, IDC_BONE_NAME, WM_SETTEXT, 0, (LPARAM) (pBone->GetName()));

	TCHAR buf[100];

	_snprintf(buf, sizeof(buf) / sizeof(TCHAR), "%g", pBone->p1);
	SendDlgItemMessage(hWnd, IDC_P1, WM_SETTEXT, 0, (LPARAM) buf);

	_snprintf(buf, sizeof(buf) / sizeof(TCHAR), "%g", pBone->p0);
	SendDlgItemMessage(hWnd, IDC_P0, WM_SETTEXT, 0, (LPARAM) buf);

	_snprintf(buf, sizeof(buf) / sizeof(TCHAR), "%g", pBone->q0);
	SendDlgItemMessage(hWnd, IDC_Q0, WM_SETTEXT, 0, (LPARAM) buf);

	_snprintf(buf, sizeof(buf) / sizeof(TCHAR), "%g", pBone->q1);
	SendDlgItemMessage(hWnd, IDC_Q1, WM_SETTEXT, 0, (LPARAM) buf);
}

static bool BoneParam_Verify(HWND hWnd)
{
	float p1, p0, q0, q1;

	if (EditCtrl_GetFloat(hWnd, IDC_P1, p1) &&
		EditCtrl_GetFloat(hWnd, IDC_P0, p0) &&
		EditCtrl_GetFloat(hWnd, IDC_Q0, q0) &&
		EditCtrl_GetFloat(hWnd, IDC_Q1, q1)) {
		if (p1 <= p0 &&
			p0 < q0 &&
			q0 <= q1) {
			BoneInfo* pBone = (BoneInfo*) GetWindowLong(hWnd,  GWL_USERDATA);
			AxAssert(pBone != NULL);

			pBone->p1 = p1;
			pBone->p0 = p0;
			pBone->q0 = q0;
			pBone->q1 = q1;

			return true;
		}

		MessageBox(hWnd, __T("Invalid values for bone parameters"), __T("Error"), MB_OK | MB_ICONSTOP); 
	}

	return false;
}

BOOL CALLBACK BoneParamDialogProc	(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch (msg) {
		case WM_INITDIALOG:
			BoneParam_OnInitDialog(hWnd, lParam);
			return TRUE;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDOK:
					if (!BoneParam_Verify(hWnd))
						break;

				case IDCANCEL:
					EndDialog(hWnd, LOWORD(wParam));
					break;
			}
			return TRUE;
	}

	return FALSE;
}

// ---------------------------------------------------------------------------
//
// Auto tendon dialog box
//

static void AutoTendon_OnInitDialog(HWND hWnd, LPARAM lParam)
{
	CenterWindow(hWnd, GetParent(hWnd));

	BoneInfo* pBone0 = KvBones::g_pCurr->m_pBoneList + lParam;
	BoneInfo* pBone1 = KvBones::g_pCurr->m_pBoneList + lParam + 1;

	SetWindowLong(hWnd,  GWL_USERDATA, lParam);

	TCHAR buf[200];

	SendDlgItemMessage(hWnd, IDC_EXT_BONE0_STR, WM_GETTEXT, sizeof(buf) / sizeof(TCHAR), (LPARAM) buf);
	strcat(buf, pBone0->GetName());
	SendDlgItemMessage(hWnd, IDC_EXT_BONE0_STR, WM_SETTEXT, 0, (LPARAM) buf);

	SendDlgItemMessage(hWnd, IDC_EXT_BONE1_STR, WM_GETTEXT, sizeof(buf) / sizeof(TCHAR), (LPARAM) buf);
	strcat(buf, pBone1->GetName());
	SendDlgItemMessage(hWnd, IDC_EXT_BONE1_STR, WM_SETTEXT, 0, (LPARAM) buf);
}

static bool AutoTendon_Verify(HWND hWnd)
{
	float f0, f1;

	if (EditCtrl_GetFloat(hWnd, IDC_EXT_BONE0_NUM, f0) &&
		EditCtrl_GetFloat(hWnd, IDC_EXT_BONE1_NUM, f1)) {
		if (f0 >= 0.0f && f0 <= 1.0f &&
			f1 >= 0.0f && f1 <= 1.0f) {
			int nBone = GetWindowLong(hWnd,  GWL_USERDATA);

			BoneInfo* pBone0 = KvBones::g_pCurr->m_pBoneList + nBone;
			BoneInfo* pBone1 = KvBones::g_pCurr->m_pBoneList + nBone + 1;

			pBone0->q0 = 1.0f - f0;
			pBone0->q1 = 1.0f + f1 * pBone0->qmplen / pBone1->qmplen;

			pBone1->p1 = 0.0f - f0 * pBone1->qmplen / pBone0->qmplen;
			pBone1->p0 = 0.0f + f1;

			const float snap = 100.0f;

			pBone0->q0 = (int)(pBone0->q0 * snap) / snap;
			pBone0->q1 = (int)(pBone0->q1 * snap) / snap;
			pBone1->p1 = (int)(pBone1->p1 * snap) / snap;
			pBone1->p0 = (int)(pBone1->p0 * snap) / snap;

			return true;
		} else {
			MessageBox(hWnd, __T("Invalid parameters: must be from 0 to 1"), __T("Error"), MB_OK | MB_ICONSTOP);
		}
	}

	return false;
}

BOOL CALLBACK AutoTendonDialogProc	(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch (msg) {
		case WM_INITDIALOG:
			AutoTendon_OnInitDialog(hWnd, lParam);
			return TRUE;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDOK:
					if (!AutoTendon_Verify(hWnd))
						break;

				case IDCANCEL:
					EndDialog(hWnd, LOWORD(wParam));
					break;
			}
			return TRUE;
	}

	return FALSE;
}

#endif

// ---------------------------------------------------------------------------
//
// BoneCurveDialogProc
//

static BoneCurveWin g_win;
static bool g_win_init;

struct WinParam
{
	UINT	nEditID;
	UINT	nSpinID;
	UINT	nTextID;
	float*	value;
};

static WinParam g_win_param[] = 
{
	{ IDC_X_MIN_EDIT, IDC_X_MIN_SPIN, IDC_X_MIN_LABEL, &g_win.minx },
	{ IDC_X_MAX_EDIT, IDC_X_MAX_SPIN, IDC_X_MAX_LABEL, &g_win.maxx },
	{ IDC_Y_MIN_EDIT, IDC_Y_MIN_SPIN, IDC_Y_MIN_LABEL, &g_win.miny },
	{ IDC_Y_MAX_EDIT, IDC_Y_MAX_SPIN, IDC_Y_MAX_LABEL, &g_win.maxy }
};

enum { kMargL = 7, kMargR = 7, kMargT = 5, kMargB = 5 };
enum { kCross = 3 };

class BoneCurveContext
{
public:
	BoneCurveContext(HWND hWnd);

	int	f2ix(float x);
	int	f2iy(float y);

	float i2fx(int x);
	float i2fy(int y);

	HWND hWndCurve;
	RECT rc;
	float scalex, scaley;
};

BoneCurveContext::BoneCurveContext(HWND hWnd)
{
	hWndCurve = GetDlgItem(hWnd, IDC_CURVE);
	
	GetClientRect(hWndCurve, &rc);
	MapWindowPoints(hWndCurve, hWnd, (LPPOINT) &rc, 2);
	InflateRect(&rc, -1, -1);

	scalex = (rc.right - rc.left - kMargL - kMargR) / (g_win.maxx - g_win.minx);
	scaley = (rc.bottom - rc.top - kMargT - kMargB) / (g_win.maxy - g_win.miny);
}

int	BoneCurveContext::f2ix(float x)
{
	return (int) (rc.left + kMargL + (x - g_win.minx) * scalex);
}

int	BoneCurveContext::f2iy(float y)
{
	return (int) (rc.top + kMargT + (g_win.maxy - y) * scaley);
}

float BoneCurveContext::i2fx(int x)
{
	return (x - rc.left - kMargL) / scalex + g_win.minx;
}

float BoneCurveContext::i2fy(int y)
{
	return g_win.maxy - (y - rc.top - kMargT) / scaley;
}

static void BoneCurve_RedrawCurve(HWND hWnd)
{
	HWND hWndCurve = GetDlgItem(hWnd, IDC_CURVE);
	
	RECT rc;
	GetClientRect(hWndCurve, &rc);
	MapWindowPoints(hWndCurve, hWnd, (LPPOINT) &rc, 2);

	InvalidateRect(hWnd, &rc, FALSE);
}

static void BoneCurve_OnInitDialog(HWND hWnd, LPARAM lParam)
{
	CenterWindow(hWnd, GetParent(hWnd));

	BoneCurve* curve = (BoneCurve*) lParam;
	SetWindowLong(hWnd,  GWL_USERDATA, lParam);

	if (!g_win_init) {
		g_win_init = true;

		g_win.minx = -0.25f;
		g_win.maxx = 1.25f;
		g_win.miny = 0.0f;
		g_win.maxy = 1.1f;
	}

	for (int i = 0; i < 4; ++ i) {

		HWND hWndSpin = GetDlgItem(hWnd, g_win_param[i].nSpinID);
		HWND hWndEdit = GetDlgItem(hWnd, g_win_param[i].nEditID);
		HWND hWndText = GetDlgItem(hWnd, g_win_param[i].nTextID);

		float value	  = * g_win_param[i].value;

		TSTR v = FormatUniverseValue(value);
		SetWindowText(hWndText, v);

		ISpinnerControl*	iS = GetISpinner(hWndSpin);
		ICustEdit*			iE = GetICustEdit(hWndEdit);

		iS->SetLimits(-5.0f, 6.0f);
		iS->SetAutoScale();
		iS->LinkToEdit(hWndEdit, EDITTYPE_FLOAT);

		iS->SetValue(value, false);
		iE->SetText(value);

		ReleaseISpinner(iS);
		ReleaseICustEdit(iE);
	}

	if (curve->kind == BoneCurve::kLinear)
		SendDlgItemMessage(hWnd, IDC_TYPE_LINEAR, BM_SETCHECK, true, 0);
	else
		SendDlgItemMessage(hWnd, IDC_TYPE_SMOOTH, BM_SETCHECK, true, 0);
}

static void BoneCurve_OnPaint(HWND hWnd)
{
	PAINTSTRUCT ps;
	HDC hDC = BeginPaint(hWnd, &ps);

	BoneCurveContext ctx(hWnd);

	// paint the background

	PatBlt(hDC, ctx.rc.left, ctx.rc.top, ctx.rc.right - ctx.rc.left, ctx.rc.bottom - ctx.rc.top, WHITENESS);

	// clip to the rect

	HRGN hRgn = CreateRectRgnIndirect(&ctx.rc);
	SelectClipRgn(hDC, hRgn); 
 
	// paint the grid

	{
		HPEN hLinPen = CreatePen(PS_DOT, 1, RGB(0x00, 0x00, 0xFF));
		HPEN hOldPen = (HPEN) SelectObject(hDC, hLinPen);

		for (float x = (float) ceil(g_win.minx); x <= floor(g_win.maxx); ++ x) {
			int ix = ctx.f2ix(x);
			MoveToEx(hDC, ix, ctx.rc.top, NULL);	LineTo(hDC, ix, ctx.rc.bottom);
		}

		for (float y = (float) ceil(g_win.miny); y <= floor(g_win.maxy); ++ y) {
			int iy = ctx.f2iy(y);
			MoveToEx(hDC, ctx.rc.left, iy, NULL);	LineTo(hDC, ctx.rc.right, iy);
		}

		SelectObject(hDC, hOldPen);
		DeleteObject(hLinPen);
	}

	// paint the curve

	BoneCurve* curve = (BoneCurve*) GetWindowLong(hWnd,  GWL_USERDATA);

	Point2 pt2[100];
	int count = curve->GenerateCurve(pt2, sizeof(pt2) / sizeof(pt2[0]));
	if (count != 0) {
		POINT pt[100];
		for (int i = 0; i < count; ++ i) {
			pt[i].x = ctx.f2ix(pt2[i].x);
			pt[i].y = ctx.f2iy(pt2[i].y);
		}
		Polyline(hDC, pt, count); 
 	}

	// paint the control points

	{
		HPEN hCtlPen = CreatePen(PS_SOLID, 1, RGB(0xFF, 0x00, 0x00));
		HPEN hOldPen = (HPEN) SelectObject(hDC, hCtlPen);

		for (int i = 0; i < 4; ++ i) {
			int x = ctx.f2ix(curve->pts[i].x);
			int y = ctx.f2iy(curve->pts[i].y);

			MoveToEx(hDC, x - kCross, y - kCross, NULL);	LineTo(hDC, x + kCross, y + kCross);
			MoveToEx(hDC, x + kCross, y - kCross, NULL);	LineTo(hDC, x - kCross, y + kCross);
		}

		SelectObject(hDC, hOldPen);
		DeleteObject(hCtlPen);
	}

	DeleteObject(hRgn);

	EndPaint(hWnd, &ps);
}

static BOOL BoneCurve_OnSetCursor(HWND hWnd, WPARAM wParam, LPARAM lParam)
{
	// get mouse pos

	BoneCurveContext ctx(hWnd);

	POINT pt;
	GetCursorPos(&pt);
	ScreenToClient(hWnd, &pt);

	// hit test control points

	BoneCurve* curve = (BoneCurve*) GetWindowLong(hWnd,  GWL_USERDATA);

	for (int i = 0; i < 4; ++ i) {
		int x = ctx.f2ix(curve->pts[i].x);
		int y = ctx.f2iy(curve->pts[i].y);

		if (pt.x - kCross <= x && x <= pt.x + kCross &&
			pt.y - kCross <= y && y <= pt.y + kCross) {
			SetCursor(LoadCursor(NULL, IDC_SIZEALL));
			return TRUE;
		}
	}

	SetCursor(LoadCursor(NULL, IDC_ARROW));
	return TRUE;
}

static BOOL BoneCurve_OnLButtonDown(HWND hWnd, WPARAM wParam, LPARAM lParam)
{
	BoneCurveContext ctx(hWnd);

	int mx = LOWORD(lParam);
	int my = HIWORD(lParam);

	// hit test control points

	BoneCurve* curve = (BoneCurve*) GetWindowLong(hWnd,  GWL_USERDATA);

	for (int i = 0; i < 4; ++ i) {
		int x = ctx.f2ix(curve->pts[i].x);
		int y = ctx.f2iy(curve->pts[i].y);

		if (mx - kCross <= x && x <= mx + kCross &&
			my - kCross <= y && y <= my + kCross) {

			int dx = x - mx;
			int dy = y - my;

			SetCapture(hWnd);

			HDC hDC = GetDC(hWnd);

			HPEN hCtlPen = CreatePen(PS_SOLID, 1, RGB(0x00, 0xFF, 0x00));
			HPEN hOldPen = (HPEN) SelectObject(hDC, hCtlPen);

			MSG msg;
			while (GetCapture() == hWnd && GetMessage(&msg, hWnd, 0, 0)) {
				switch (msg.message) {

					case WM_KEYDOWN:
						if (lParam == VK_ESCAPE)
							goto bail;
						break;

					case WM_LBUTTONUP:
						goto good;

					case WM_RBUTTONDOWN:
						goto bail;

					case WM_MOUSEMOVE:
						mx = LOWORD(msg.lParam);
						my = HIWORD(msg.lParam);

						// MoveToEx(hDC, x - kCross, y - kCross, NULL);	LineTo(hDC, x + kCross, y + kCross);
						// MoveToEx(hDC, x + kCross, y - kCross, NULL);	LineTo(hDC, x - kCross, y + kCross);

						break;

					default:
						DispatchMessage(&msg);
						break;
				}
			}

good:

			curve->pts[i].x = ctx.i2fx(mx + dx);
			curve->pts[i].y = ctx.i2fy(my + dy);

			// disallow crossovers

			if (i > 0 && curve->pts[i].x < curve->pts[i - 1].x)
				curve->pts[i].x = curve->pts[i - 1].x;
			else if (i < 3 && curve->pts[i].x > curve->pts[i + 1].x)
				curve->pts[i].x = curve->pts[i + 1].x;

			BoneCurve_RedrawCurve(hWnd);

bail:

			SelectObject(hDC, hOldPen);
			DeleteObject(hCtlPen);
			ReleaseDC(hWnd, hDC);

			ReleaseCapture();

			break;
		}
	}

	return TRUE;
}

static void BoneCurve_OnCurveType(HWND hWnd, UINT nID)
{
	BoneCurve* curve = (BoneCurve*) GetWindowLong(hWnd,  GWL_USERDATA);

	switch (nID) {
		case IDC_TYPE_LINEAR:
			if (curve->kind != BoneCurve::kLinear) {
				curve->kind =  BoneCurve::kLinear;
				BoneCurve_RedrawCurve(hWnd);
			}
			break;

		case IDC_TYPE_SMOOTH:
			if (curve->kind != BoneCurve::kSmooth) {
				curve->kind =  BoneCurve::kSmooth;
				BoneCurve_RedrawCurve(hWnd);
			}
			break;
	}
}

static void BoneCurve_OnSpinnerChange(HWND hWnd, WPARAM wParam, LPARAM lParam)
{
	for (int i = 0; i < 4; ++ i) {
		if (g_win_param[i].nSpinID == LOWORD(wParam)) {

			// get new value

			ISpinnerControl* iS = GetISpinner(GetDlgItem(hWnd, g_win_param[i].nSpinID));
			*g_win_param[i].value = iS->GetFVal();
			ReleaseISpinner(iS);

			// redraw label

			TSTR v = FormatUniverseValue(*g_win_param[i].value);
			SetWindowText(GetDlgItem(hWnd, g_win_param[i].nTextID), v);

			// redraw the curve

			BoneCurve_RedrawCurve(hWnd);

			break;
		}
	}
}

BOOL CALLBACK BoneCurveDialogProc	(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg) {
		case WM_INITDIALOG:
			BoneCurve_OnInitDialog(hWnd, lParam);
			return TRUE;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDOK:
				case IDCANCEL:
					EndDialog(hWnd, LOWORD(wParam));
					break;

				case IDC_TYPE_LINEAR:
				case IDC_TYPE_SMOOTH:
					BoneCurve_OnCurveType(hWnd, LOWORD(wParam));
					break;
			}
			return TRUE;

		case CC_SPINNER_CHANGE:
			BoneCurve_OnSpinnerChange(hWnd, wParam, lParam);
			break;

		case WM_PAINT:
			BoneCurve_OnPaint(hWnd);
			break;

		case WM_SETCURSOR:
			return BoneCurve_OnSetCursor(hWnd, wParam, lParam);

		case WM_LBUTTONDOWN:
			return BoneCurve_OnLButtonDown(hWnd, wParam, lParam);
	}

	return FALSE;
}
