//
//	Copyright (C) 1998 Kostya Vasilyev, kostik@pobox.com
//
//	This program is free software; you can redistribute it and/or
//	modify it under the terms of the GNU General Public License
//	as published by the Free Software Foundation; either version 2
//	of the License, or (at your option) any later version.
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

#include "KvBonesUndo.h"
#include "KvBones.h"

// ---------------------------------------------------------------------------
//
// UndoVertexEdit
//

int UndoVertexEdit::g_count = 0;

UndoVertexEdit::~UndoVertexEdit()
{
}

void UndoVertexEdit::Restore(int isUndo)
{
	KvLocalModData* mod = m_bones->m_pModData;
	if (m_vertex < mod->m_nCount) {
		Vtx& vtx = mod->m_pList[m_vertex];

		VtxState s;
		vtx.SaveState(m_bones->m_pBoneList, m_bones->m_nBoneCount, s);
		vtx.RestoreState(m_bones->m_pBoneList, m_bones->m_nBoneCount, m_save);
		m_save = s;

		KvBonesDlgProc::instance.BindToVertex(-1, -1, 0);
		KvBonesDlgProc::instance.ShowVertexInfo(m_vertex);

		m_bones->ForceRedraw();
	}
}

void UndoVertexEdit::Redo()
{
	Restore(true);
}

int UndoVertexEdit::Size()
{
	return sizeof(*this);
}

void UndoVertexEdit::Begin(KvBones* bones, int nVertex)
{
	if (!theHold.Holding())
		theHold.Begin();

	++ g_count;
	theHold.Put(new UndoVertexEdit(bones, nVertex));
}

void UndoVertexEdit::End()
{
	if (g_count != 0) {
		g_count = 0;
		theHold.Accept(_T("Edit Vertex"));
		theHold.End();
	}
}

UndoVertexEdit::UndoVertexEdit(KvBones* bones, int nVertex) :
	m_bones(bones), m_vertex(nVertex)
{
	KvLocalModData* mod = m_bones->m_pModData;
	if (m_vertex < mod->m_nCount) {
		Vtx& vtx = mod->m_pList[m_vertex];
		vtx.SaveState(m_bones->m_pBoneList, m_bones->m_nBoneCount, m_save);
	}
}

// ---------------------------------------------------------------------------
//
// UndoBoneEdit
//

UndoBoneEdit::~UndoBoneEdit()
{
	delete[] m_state;
}

void UndoBoneEdit::Restore(int isUndo)
{
	if (m_count == m_bones->m_nBoneCount) {
		// bones
		for (int i = 0; i < m_count; ++ i) {
			BoneState s;
			m_bones->m_pBoneList[i].SaveState(s);
			m_bones->m_pBoneList[i].RestoreState(m_state[i]);
			m_state[i] = s;
		}
		// mod state
		KvLocalModDataState m;
		m_bones->m_pModData->SaveState(m);
		m_bones->m_pModData->RestoreState(m_mod);
		m_mod = m;
		// redraw
		m_bones->ForceRedraw();
	}
}

void UndoBoneEdit::Redo()
{
	Restore(true);
}

int UndoBoneEdit::Size()
{
	return sizeof(*this) + m_count * sizeof(BoneState) + m_mod.m_nCount * sizeof(Vtx);
}

UndoBoneEdit* UndoBoneEdit::Create(KvBones* bones)
{
	if (!theHold.Holding())
		theHold.Begin();

	UndoBoneEdit* p = new UndoBoneEdit(bones);
	theHold.Put(p);
	return p;
}

UndoBoneEdit::UndoBoneEdit(KvBones* bones) : m_bones(bones), m_count(bones->m_nBoneCount)
{
	m_state = new BoneState[m_count];

	// bones
	for (int i = 0; i < m_count; ++ i) {
		m_bones->m_pBoneList[i].SaveState(m_state[i]);
	}
	// mod state
	m_bones->m_pModData->SaveState(m_mod);
}
