#pragma once

//
//	Copyright (C) 1998 Kostya Vasilyev, kostik@pobox.com
//
//	This program is free software; you can redistribute it and/or
//	modify it under the terms of the GNU General Public License
//	as published by the Free Software Foundation; either version 2
//	of the License, or (at your option) any later version.
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**********************************************************************
 *<
	FILE: kvBones.h

	DESCRIPTION:	Template Utility

	CREATED BY:

	HISTORY:

 *>	Copyright (c) 1997, All Rights Reserved.
 **********************************************************************/

#include "KvBonesDefs.h"
#include "KvBonesData.h"

#include "iparamm.h"
#include "Simpmod.h"
#include "Simpobj.h"

#include "meshadj.h"

TCHAR *GetString(int id);

extern HINSTANCE hInstance;

// ---------------------------------------------------------------------------

class	KvLocalModData;

struct	BoneInfo;
struct	Vtx;

// ---------------------------------------------------------------------------

#define KVBONESOBJ_CLASS_ID	Class_ID(0x671b5fd2, 0x9303c5b5)
#define KVBONES_CLASS_ID	Class_ID(0x671b5fd2, 0x9303c5b6)

// ---------------------------------------------------------------------------
//
// class KvBonesObject
//

class KvBonesObject : public SimpleWSMObject
{
public:
	KvBonesObject();
	virtual ~KvBonesObject();

	// From Animatable		
	void DeleteThis() {delete this;}		
	void BeginEditParams( IObjParam  *ip, ULONG flags,Animatable *prev );
	void EndEditParams( IObjParam *ip, ULONG flags,Animatable *next);

	Class_ID ClassID() {return KVBONESOBJ_CLASS_ID;}

	// from object		
	int DoOwnSelectHilite() { return TRUE;}
	CreateMouseCallBack* GetCreateMouseCallBack();

	// From WSMObject
	Modifier *CreateWSMMod(INode *node);
	void BuildMesh(TimeValue t);
};

// ---------------------------------------------------------------------------
//
// class KvBonesObjectCreateCallback
//

class KvBonesObjectCreateCallback : public CreateMouseCallBack
{
	KvBonesObject *ob;	
	Point3 p0, p1;
	IPoint2 sp0, sp1;

public:
	int proc( ViewExp *vpt,int msg, int point, int flags, IPoint2 m, Matrix3& mat);
	void SetObj(KvBonesObject *obj) { ob = obj; }
};

// ---------------------------------------------------------------------------
//
// class KvBones
//

class KvBones : public SimpleWSMMod
{
public:

	typedef SimpleWSMMod inherited;

	static IObjParam *	ip;			//Access to the interface
	static HWND			hWndDlg;
	static KvBones*		g_pCurr;
	//Class vars
	static IParamMap *	pmapParam;	//handle to the parammap		

	//Constructor/Destructor
	KvBones();
	KvBones(KvBonesObject* ob, INode* node);
	~KvBones();

	// From Animatable
	void DeleteThis() { delete this; }
	void GetClassName(TSTR& s) { s= TSTR(GetString(IDS_CLASS_NAME)); }  
	virtual Class_ID ClassID() { return KVBONES_CLASS_ID;}	
	virtual SClass_ID SuperClassID() { return WSM_CLASS_ID; }
	RefTargetHandle Clone(RemapDir& remap = NoRemap());
	TCHAR *GetObjectName() { return GetString(IDS_CLASS_NAME); }

	IOResult Save(ISave *isave);
	IOResult Load(ILoad *iload);

	//From Modifier
	//TODO: Add the channels that the modifier needs to perform its modification
	ChannelMask ChannelsUsed()  { return PART_GEOM|PART_TOPO|PART_VERTCOLOR|PART_TOPO; }
	//TODO: Add the channels that the modifier actually modifies
	ChannelMask ChannelsChanged() { return GEOM_CHANNEL|VERTCOLOR_CHANNEL|TOPO_CHANNEL; }
	void ModifyObject(TimeValue t, ModContext &mc, ObjectState *os, INode *node);

	//TODO: Return the ClassID of the object that the modifier can modify
	Class_ID InputType() {return defObjectClassID;}
	BOOL DependOnTopology(ModContext &mc) { return TRUE; }
	Interval LocalValidity(TimeValue t);

	// From BaseObject
	//TODO: Return true if the modifier changes topology
	BOOL ChangeTopology() {return FALSE;}
	int GetParamBlockIndex(int id) {return id;}

	//From ReferenceMaker
	//TODO :Implement the following methods for references		
	int NumRefs();
	RefTargetHandle GetReference(int i);
	void SetReference(int i, RefTargetHandle rtarg);
	
	int NumSubs() { return 1; }  
	Animatable* SubAnim(int i) { return pblock; }
	TSTR SubAnimName(int i) { return GetString(IDS_PARAMS);}

	// From SimpleMod
	Deformer& GetDeformer(TimeValue t,ModContext &mc,Matrix3& mat,Matrix3& invmat);

	RefResult NotifyRefChanged( Interval changeInt,RefTargetHandle hTarget, 
	   PartID& partID, RefMessage message);
	
	CreateMouseCallBack* GetCreateMouseCallBack() {return NULL;}
	void BeginEditParams(IObjParam *ip, ULONG flags,Animatable *prev);
	void EndEditParams(IObjParam *ip, ULONG flags,Animatable *next);

	Interval GetValidity(TimeValue t);
	ParamDimension *GetParameterDim(int pbIndex);
	TSTR GetParameterName(int pbIndex);

	// Automatic texture support
	BOOL HasUVW();
	void SetGenUVW(BOOL sw);

	// ----- KV bones code

	void				SetBoneList(TimeValue t, INode** pList, int nCount);
	void				UpdateBoneData(TimeValue t);
	KvLocalModData*		BuildBoneVertexMapping(Mesh* mesh);
	void				ForceUpdate();
	void				ForceRedraw();

	void				OnCopyPos(TimeValue t);
	void				OnPastePos(TimeValue t);

	int					m_nBoneCount;
	BoneInfo*			m_pBoneList;
	int					m_ModDataSeed;
	int					m_nSelectBone;	// or -1 if none

	int					m_nFixedRefs;
	RefTargetHandle		m_pFixedRefs[1];

	Mesh				m_meshCopy;
	KvLocalModData*		m_pModData;

	BonePosBuffer		m_PosBuffer;

	static BoneInfo*	s_bonelist_begin();
	static BoneInfo*	s_bonelist_end();
};

// ---------------------------------------------------------------------------
//
// class KvBonesMouseCallback
//

class KvBonesMouseCallback : public MouseCallBack
{
public:
	virtual int proc( 
		HWND hwnd, 
		int msg, 
		int point, 
		int flags, 
		IPoint2 m );

	KvBones*					bones;

	static KvBonesMouseCallback instance;

private:
	HCURSOR	hcurArrow;
	HCURSOR	hcurPickWeight;
	HCURSOR	hcurAddVertex;
	HCURSOR hcurRemoveVertex;
	HCURSOR hcurSetWeight;
	HCURSOR	hcurSelect;

	IPoint2	mpt[2];
	bool    selectRect;

	void InvertSelectRect(HWND hwnd);
	int	HitTest(HWND hwnd, HitRegion& hr, SubObjHitList& hitList);
};

// ---------------------------------------------------------------------------
//
// class KvBonesCommandMode
//

#define CID_BONESCOMMAND	CID_USER + 1

class KvBonesCommandMode : public CommandMode
{
public:
	MouseCallBack*	proc;

	int Class();
	int ID();
	MouseCallBack *MouseProc(int *numPoints);
	ChangeForegroundCallback *ChangeFGProc();
	BOOL ChangeFG( CommandMode *oldMode );
	void EnterMode();
	void ExitMode();

	void Install(IObjParam* ip, KvBones* bones, KvBonesMouseCallback& mc);
	void Remove(IObjParam* ip);

	static KvBonesCommandMode instance;
};

// ---------------------------------------------------------------------------
//
// class KvBonesDlgProc
//

class KvBonesDlgProc : public ParamMapUserDlgProc
{
public:
	BOOL DlgProc(TimeValue t,IParamMap *map,HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam);
	void DeleteThis();

	static KvBonesDlgProc instance;

	float	GetWeight();
	void	BindToVertex(int nVertex, int nBone, float wght);
	void	ShowVertexInfo(int n);

	void	SetVertexWeight(Vtx& vtx, int nBone);

private:
	void OnInitDialog(HWND hWnd);

	HWND	m_hWnd;

	HWND	m_hwndWeightSpin;
	HWND	m_hwndWeightEdit;
	HWND	m_hwndWeightText;

	HWND	m_hwndInfoV;
	HWND	m_hwndInfo0;
	HWND	m_hwndInfo1;

	int		m_nBoundVertex;
	int		m_nBoundBone;
};

// ---------------------------------------------------------------------------
//
// class KvSelectCallback
//

class KvSelectCallback : public HitByNameDlgCallback
{
public:

	KvSelectCallback(TimeValue t, KvBones* pBones);

	virtual TCHAR *dialogTitle();
	virtual TCHAR *buttonText();

	virtual BOOL useProc();
	virtual void proc(INodeTab &nodeTab);

	virtual BOOL useFilter();
	virtual int filter(INode *node);

private:
	TimeValue	m_t;
	KvBones*	m_pBones;
};

// ---------------------------------------------------------------------------
//
// This is the callback object used by modifiers to deform "Deformable" objects.
//

class KvBonesDeformer : public Deformer
{
public:
	KvBonesDeformer();
	Point3 Map(int i, Point3 p); 

	void		Init(TimeValue t, KvBones* pBones, float value, KvLocalModData* modData);

private:
	KvBones*			m_pBones;
	float				m_Value;

	KvLocalModData*		m_pModData;
};

// ---------------------------------------------------------------------------
//
// KvBonesObjClassDesc
//

class KvBonesObjClassDesc : public ClassDesc
{
	public:
	int 			IsPublic() {return 1;}
	void *			Create(BOOL loading = FALSE) {return new KvBonesObject();}
	const TCHAR *	ClassName() {return GetString(IDS_OBJ_CLASS_NAME);}
	SClass_ID		SuperClassID() {return WSM_OBJECT_CLASS_ID;}
	Class_ID		ClassID() {return KVBONESOBJ_CLASS_ID;}
	const TCHAR* 	Category() {return GetString(IDS_CATEGORY);}
	void			ResetClassParams (BOOL fileReset);
};

// ---------------------------------------------------------------------------
//
// KvBonesClassDesc
//

class KvBonesClassDesc : public ClassDesc
{
	public:
	int 			IsPublic() {return 0;}
	void *			Create(BOOL loading = FALSE) {return new KvBones();}
	const TCHAR *	ClassName() {return GetString(IDS_WSM_CLASS_NAME);}
	SClass_ID		SuperClassID() {return WSM_CLASS_ID;}
	Class_ID		ClassID() {return KVBONES_CLASS_ID;}
	const TCHAR* 	Category() {return GetString(IDS_CATEGORY);}
	void			ResetClassParams (BOOL fileReset);
};

// ---------------------------------------------------------------------------
//
// Parameter block values
//

#define	PB_VAL_VALUE			0
#define	BONES_DATA_CHUNK		1004

